<el-date-picker
  :type="data.{{{widget.widget_id}}}.type"
  :placeholder="data.{{{widget.widget_id}}}.placeholder"
  :size="data.{{{widget.widget_id}}}.size"
  :format="data.{{{widget.widget_id}}}.format"
  :readonly="data.{{{widget.widget_id}}}.readonly"
  :disabled="data.{{{widget.widget_id}}}.disabled"
  :editable="data.{{{widget.widget_id}}}.editable"
  :clearable="data.{{{widget.widget_id}}}.clearable"
  v-model="state.{{{widget.widget_id}}}.value"
  {% if widget._changes_handled == true %}
        @change="post('/{{{widget.widget_id}}}/value_changed')"
  {% endif %}
>
</el-date-picker>
